#include <Arduino.h>
#include <RotaryEncoder.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Wire.h>
#include "ESP8266WiFi.h"
#include <WiFiUdp.h>
#include <NTPClient.h>
#include <EEPROM.h>
#define EEPROM_SIZE 12


extern "C" {
   #include "gpio.h"
 }


#define SCREEN_WIDTH 128 // OLED display width,  in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

// declare an SSD1306 display object connected to I2C
Adafruit_SSD1306 oled(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, -1);

// Example for ESP8266 NodeMCU with input signals on pin D5 and D6
#define PIN_IN1 D5
#define PIN_IN2 D6
#define PIN_IN3 D2
#define PIN_VIBRATOR D1

// A pointer to the dynamic created rotary encoder instance.
// This will be done in setup()
RotaryEncoder *encoder = nullptr;
/**
 * @brief The interrupt service routine will be called on any change of one of the input signals.
 */
IRAM_ATTR void checkPosition(){
  encoder->tick(); // just call tick() to check the state.
}


const char *ssid     = "Wifi";        //Enter the ssid of your wifi
const char *password = "Password";         //Enter the password of your wifi
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, "north-america.pool.ntp.org");     // choose server to get the time from, ideally choose the ones closer to your region


int ledState = LOW;             // ledState used to set the LED

int buttonState = LOW;


unsigned long currentMillis = 0;
unsigned long previousActivityMillis = 0;
unsigned long tempIntMillis  = 0;
const long debounceInterval= 500;
const long interval = 50;

static int pos = 0;

char timebuffer[40];


int timeMins= 0;
int alarmSet= 0;
int selectMenu = 0;
int state = 0;
int nextState= 0;
const int alarmNumber = 5;
int upperLimit[]= {3,alarmNumber,2,9,1,1};      //upperlimit for scroll wheel in different menus
int alarmTime[alarmNumber] = {0,0,0};   //time of alarm stored in minutes
int selectAlarm= 0;   // index for scrolling digits to set the alarm
int sensorValueVolt= 0; 
int whichAlarm= 0;   // index for which alarm is being modified 
int ringAlarm= 0;    // index for which alarm is ringing currently



IRAM_ATTR void checkSwitch()             //Interrupt function for whenever the switch is closed 
{ 
  currentMillis = millis();
  previousActivityMillis = currentMillis;
  if (currentMillis - tempIntMillis > debounceInterval) {         //this is a software implemented debounce to to avoid false presses
    tempIntMillis = currentMillis;
    buttonState = HIGH;               // buttonstate is the variable that the rest of the program reads as indication that switch has been pressed
  }
}



void setup()
{ 
  EEPROM.begin(512);
  pinMode(D7, OUTPUT);   //declare d7 as an output   
  pinMode(PIN_IN3, INPUT_PULLUP);      //Connects D2 to pullup
  pinMode(PIN_VIBRATOR, OUTPUT);
  digitalWrite(D7, LOW);    // this "turns off" the mosfet gate that controls the connection between switch and reset

  Serial.begin(115200);

  encoder = new RotaryEncoder(PIN_IN1, PIN_IN2, RotaryEncoder::LatchMode::TWO03);   //sets up the rotary encoder 

  attachInterrupt(digitalPinToInterrupt(PIN_IN1), checkPosition, CHANGE);
  attachInterrupt(digitalPinToInterrupt(PIN_IN2), checkPosition, CHANGE);
  attachInterrupt(digitalPinToInterrupt(PIN_IN3), checkSwitch, FALLING);

  initWifi();      //function to start the wifi connection


  timeClient.begin();
  timeClient.update();   // first few updates seemed to be buggy

  // initialize OLED display with address 0x3C for 128x64
  Wire.begin(D3,D4); //D3 as SDA and D4 as SCL   instead of the D1 Mini's default d2 and d1
  if (!oled.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {
    Serial.println(F("SSD1306 allocation failed"));
    while (true);
  }

  updateAlarm();  //updates alarms from EEPROM
  
} // setup()





// Read the current position of the encoder and print out when changed.
void loop()
{
  oled.clearDisplay(); // clear display after every loop so new updated display doesnt conflict with previous displayed graphics

  if((alarmCheck()) && (state != 3) ){           //check if time of the alarm is equal to current time after alarm is set
    state = 4;
  }
  
  selectMenu= encoderSelect(upperLimit[state], selectMenu);         //function to map scrollwheel options in the menu based on the state.
  currentStateCalc();                
  nextState= nextStateCalc(state, selectMenu);
  sleepFunction();

} 



void currentStateCalc(){
switch(state) {               // same format for all the cases
      case 0 :                // first get the state, then the selection and do some action based off of it
         if(selectMenu == 0){
           displayTime(getTime(0), getTime(1), getTime(3));
         }
         else if(selectMenu == 1){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            oled.println("Alarms");
            oled.display();
         }
         else if(selectMenu ==2){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            oled.println("Battery");
            oled.display();
         }
         
         else if(selectMenu == 3){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(35, 25);        // position to display
            oled.println("Shut");
            oled.println("Down");
            oled.display();
         }

         if(nextState == 1 ){
           state = 1;
           selectMenu = 0;
         }
         else if(nextState == 5 ){
           state = 5;
           selectMenu = 0;
         }  
      
         break;
      case 1 :
        if(selectMenu == alarmNumber){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            oled.println("Exit");
            oled.display();
         }
        else{
            whichAlarm= selectMenu;
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            if(alarmTime[selectMenu] != 0){
            int minTemp = alarmTime[selectMenu] % 60;
            int hourTemp = ((alarmTime[selectMenu] / 60) >= 12) ? (alarmTime[selectMenu] / 60)%12 : (alarmTime[selectMenu] / 60);
              hourTemp = (hourTemp == 0) ? 12 : hourTemp;
              if((alarmTime[selectMenu] / 60) >= 12){
                sprintf(timebuffer, "%02d:%02d PM", hourTemp, minTemp);
              }
              else{
                sprintf(timebuffer, "%02d:%02d AM", hourTemp, minTemp);
              }
            }
            else{
              sprintf(timebuffer, "Empty %d", selectMenu);  
            }
            oled.print(timebuffer);
            oled.display();
        }
        if(nextState == 2 ){
           state = 2;
           selectMenu = 0;
         }
         else if(nextState == 0){
           state = 0;
           selectMenu = 0;
         }
         break;


      case 2 :
         if(selectMenu == 0){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            oled.println("Edit");
            oled.display();
         }
         else if(selectMenu == 1){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            oled.println("Delete");
            oled.display();
         }
         else if(selectMenu == 2){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            oled.println("Exit");
            oled.display();
         }
         if(nextState == 3 ){
           state = 3;
           selectMenu =0;
         }
         else if(nextState == 1){
           state = 1;
           selectMenu =0;
         }
         else if(nextState == 0){
           state = 0;
           selectMenu =0;
         }
         break;
      case 3 :
          alarmSetter();
           state = 1;
           selectMenu =0;

         break;
      case 4:
        { unsigned long previousMillisRing;
          int ringState = HIGH;
          while(buttonState != HIGH){
            unsigned long currentMillisRing = millis();
            if ((currentMillisRing - previousMillisRing >= 200) && (ringState == LOW)) {
              previousMillisRing = currentMillisRing;
              }
            else if((currentMillisRing - previousMillisRing >= 50) && (ringState == HIGH)) {
                ringState = LOW;
              }
            
            digitalWrite(PIN_VIBRATOR, ringState);
            displayTime(getTime(0), getTime(1), getTime(3));
            
            oled.setTextSize(1);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(10, 45);
            oled.print ("ALARM ");
            oled.print (ringAlarm);
            oled.println (" is RINGING");
          }

          digitalWrite(PIN_VIBRATOR, LOW);
          pressRead();
          alarmTime[ringAlarm] = 0;
          EEPROM.put((ringAlarm*4) -1, 0);
          EEPROM.commit();
          state = 0;
          selectMenu = 0;
        }

          break;

      case 5:
         BatteryStatus();
         while(buttonState != HIGH){
           oled.display(); 
         }

         pressRead();
         if(nextState == 0 ){
           state = 0;
           selectMenu = 0;
         } 
        
         break;

      default :
         selectMenu =0;
   }  
}



int nextStateCalc(int currentState, int selection){
  switch(currentState) {
      case 0 :
          if((selection == 1) && (buttonState == HIGH)){
            pressRead();
            return 1;
          }
          else if((selection == 2) && (buttonState == HIGH)){
            pressRead();
            sensorValueVolt = analogRead(A0); 
            return 5;
          }
          else if((selection == 3) && (buttonState == HIGH)){
            oled.setTextSize(2);          // text size
            oled.setTextColor(WHITE);     // text color
            oled.setCursor(15, 25);        // position to display
            oled.println("BYE");
            oled.display();
            delay(1500);
            oled.clearDisplay();
            oled.display();
            ESP.deepSleep(0);
          }
            else{
              return 0;
            }

         break;
      case 1 :
         if((selection == alarmNumber) && (buttonState == HIGH)){
           pressRead();
           return 0;
         }
         else if((selection != alarmNumber) && (buttonState == HIGH)){
           pressRead();
           return 2;
         }
         else{ 
           return 1;
         } 
         break;
      case 2 :
         if((selection == 0) && (buttonState == HIGH)){
           pressRead();
           return 3; 
         }
         else if((selection == 1) && (buttonState == HIGH)){
           alarmTime[whichAlarm] = 0;
           pressRead();
           return 1; 
         }
         else if((selection == 2) && (buttonState == HIGH)){
           pressRead();
           return 0;
         }
         else{
           return 2;
         }
         break;
      case 3 :
         
         break;
      case 4 :
          break;
      case 5:
          if(buttonState == HIGH){
            pressRead();
            return 0;
          }
          break;
      default :
         printf("Invalid input next state\n" );
   }  
}




void sleepFunction(){                //function to put the device into deep sleep depending on if its idle or if theres any alarms nearby
currentMillis = millis();

if(((previousActivityMillis - currentMillis) >= 60e6)){            //if its been idle for over 60 seconds
  if(nearestAlarm() > 30){         //check if nearest alarm is 30+ minutes away
    oled.clearDisplay();
    oled.display();
    delay(500);
    ESP.deepSleep(1740e6);       //go to sleep for 29 minutes      1740 is the seconds, so 29 minutes
  }
  else if((nearestAlarm() > 15)){
    oled.clearDisplay();
    oled.display();
    delay(500);
    ESP.deepSleep(840e6); 
  }
  else if((nearestAlarm() > 5)){
    oled.clearDisplay();
    oled.display();
    delay(500);
    ESP.deepSleep(240e6);
  }
  else if((nearestAlarm() > 2)){
    oled.clearDisplay();
    oled.display();
    delay(500);
    ESP.deepSleep(60e6);
  }
}

}




void BatteryStatus()
{ 
  int R1= 100000;  // inside the wemos theres a 100k and 220k v divider
  int RTotal= 790000;  // 220k and 100k resistor inside, and 470k resistor outside
  
  int pot = 0;
  int voltage = ((RTotal/R1) * (sensorValueVolt/ 1023));  
  // print out the value you read:
  if(voltage < 3.3){
    voltage = 3.3;
  }
  if(voltage > 4.2){
    voltage = 4.2;
  }
  pot= map (voltage, 3.3, 4.2 ,0,100);    // map (voltage, battery low V, Battery high V, target range min , target range max) 
    oled.setTextSize(2);          // text size
    oled.setTextColor(WHITE);     // text color
    oled.setCursor(20, 20);        // position to display
    oled.println("Battery:");
    oled.print("  ");
    oled.print(pot);
    oled.println(" %");
    oled.display();
}



void alarmSetter(){              //function to set alarm
  int alarmDigitLimit[] = {1, 9, 6, 9, 1};          // the max value for each digit of the alarm, theres 4 digits (12:59), and one more for AM or PM
  int alarmDigitVal[]= {0,0,0,0,0};                 //this array holds the value of each digit of the alarm, its initialized to 0
  int alarmDigitWeight[]= {600,60,10,1,720};        //weight of each digit of alarm in minutes,   1 hour would have weight of 60 for example
  int temp2 = alarmTime[selectMenu] % 60;           //temp1 and temp2 gets the value of the alarm you are trying to edit, so you know what they were previously
    int temp1 = ((alarmTime[selectMenu] / 60) >= 12) ? (alarmTime[selectMenu] / 60)%12 : (alarmTime[selectMenu] / 60);
      if((alarmTime[selectMenu] / 60) >= 12){
        alarmDigitVal[4] = 1;
      }
      else{
        alarmDigitVal[4] = 0;
        }
  alarmDigitVal[0] = temp1 / 10;              
  alarmDigitVal[1] = temp1 % 10;
  alarmDigitVal[2] = temp2 / 10; 
  alarmDigitVal[3] = temp2 % 10;

  alarmTime[whichAlarm] = 0;
  
  for(int i=0; i<5; i++){  //goes through all 4 digits, and ampm to let u select the alarm
    selectAlarm= 0;
    while(buttonState != HIGH){
      oled.clearDisplay();
      
      
      if((alarmDigitVal[0] == 0)&& (i != 0) && (i != 3)){               //to make sure you can not set alarm over 12:
      selectAlarm= encoderSelect(alarmDigitLimit[i], selectAlarm);
      }
      else{
        if(i== 1){
        selectAlarm= encoderSelect(2, selectAlarm);
        }
        else{
        selectAlarm= encoderSelect(alarmDigitLimit[i], selectAlarm);  
        }
      }
      if((alarmDigitVal[2] == 6)&& (i == 3)){
      selectAlarm= 0;
      }
      else if((i == 3)){
          selectAlarm= encoderSelect(alarmDigitLimit[i], selectAlarm);  
      }

      oled.setTextSize(2);          // text size
      oled.setTextColor(WHITE);     // text color
      oled.setCursor(15, 25);        // position to display
      oled.print(alarmDigitVal[0]);
      oled.print(alarmDigitVal[1]);
      oled.print(":");
      oled.print(alarmDigitVal[2]);
      oled.print(alarmDigitVal[3]);
      if(alarmDigitVal[4] != 1){
        oled.print(" AM");
      }
      else{
        oled.print(" PM");
      }
      oled.display();
      alarmDigitVal[i] = selectAlarm;
    }
    
    pressRead();
    alarmTime[whichAlarm] = alarmDigitWeight[i]*alarmDigitVal[i] + alarmTime[whichAlarm];        //converts alarm time into minutes so its easier to compare to current time
  }
  
  EEPROM.put((whichAlarm*4) -1,alarmTime[whichAlarm]);      //updates the alarm in the flash memory so that it is not erased even after restarting the device
  EEPROM.commit();
  alarmSet= 1;

}


void updateAlarm(){                 //Reads alarm settings from flash memory and updates it into the temporary memory
  for(int i=0; i<alarmNumber; i++){
     EEPROM.get((i*4) -1, alarmTime[i]);
  }
}


bool alarmCheck(){                  //checks the alarms against the current time to determine if alarm should turn on
  bool temp= false;

  for(int i=0; i< alarmNumber; i++ ){
    if (timeMinCalc() == alarmTime[i]){
      temp = true;
      ringAlarm = i;
    }
  }
  if(temp == true){
    return true;
  }
  else{
    return false;
  }
}


int nearestAlarm(){                  //checks the alarms against the current time to determine if there is an alarm soon and how much time left
  int minAlarm= 1500;   //more minutes than in a day

  for(int i=0; i< alarmNumber; i++ ){
    if (((timeMinCalc() - alarmTime[i]) > 0 && ((timeMinCalc() - alarmTime[i]) < minAlarm)) && (timeMinCalc() != 0)){
      minAlarm = timeMinCalc() - alarmTime[i];
    }
  }
  return minAlarm;
}



int encoderSelect(int upperLimit, int sel){       //translates the rotary encoder's turns into a slection, upper limit and current position
  encoder->tick();
  int newPos = encoder->getPosition();
  if(((pos - newPos) > 1) || ((pos - newPos) < -1)) {
    sel = sel+ ((int)(encoder->getDirection()));
    pos = newPos;
  }
  if(sel < 0){
    sel = 0;
  }
  if(sel > upperLimit){
    sel = upperLimit;
  }

  return sel;
}



void pressRead(){                  // switches buttonState to low to confirm button press has been read
  if (buttonState == HIGH){
    buttonState = LOW;
  }
}



void displayTime(int hr, int m, int am){         // takes the time constants and displays the time pretty nicely

  oled.setTextSize(2);          // text size
  oled.setTextColor(WHITE);     // text color
  oled.setCursor(15, 25);        // position to display
  sprintf(timebuffer, "%02d:%02d", hr, m);
  oled.print(timebuffer);       //using sprintf and buffer because its really annoying to format the strings otherwise

  if(am== 0){
  oled.println(" AM"); // text to display
  }
  else{
  oled.println(" PM"); // text to display
  }
  oled.display();               // show on OLED

}




void initWifi() {                             //code to connect to wifi
   
   WiFi.mode(WIFI_STA);
   WiFi.begin(ssid, password);           //this ssid and password value should be changed in the beginning of the code to your wifi's ssid and password
   while ((WiFi.status() != WL_CONNECTED)) {
      delay(50);
      Serial.print(".");
   }
   Serial.println("");
   Serial.print("WiFi connected, IP address: "); 
   Serial.println(WiFi.localIP());
 }





int getTime(int input){          // input 0=hour  1=min 2= seconds  3=am or pm
  timeClient.update();           // gets times from NTP client and returns hours mins seconds or am/pm         
  int hour= timeClient.getHours();
  int min= timeClient.getMinutes();
  int amPm= 0;

  if(dLSaving()){
    hour = hour -4;              // EDT is 4 hour behind UTC     
  }
  else{
    hour = hour -5;              // EST is 5 hour behind UTC
  }
  if(hour < 0){
    hour = hour +24;
  }

  if(hour < 12){
    amPm= 0;
  }
  else{
    amPm= 1;
  }

  if(hour > 12){ 
   hour= hour %12;
  }

 switch(input) {      
      case 0 :
         return hour;
         break;
      case 1 :
         return min;
         break;
      case 2 :
         return timeClient.getSeconds();
         break;
      case 3 :
         return amPm;
         break;

      default :
         printf("Invalid input get time\n" );
   }
  
}



bool dLSaving(){                  //checks the dates and everything to confirm daylight saving. Change it according to your region
  //Get a time structure
  time_t epochTime = timeClient.getEpochTime();                //gets the time that passed since some day in 1970      
  struct tm *ptm = gmtime ((time_t *)&epochTime);              //calculate GMT from the time above

  int monthDay = ptm->tm_mday;
  int currentMonth = ptm->tm_mon+1;
  int currentYear = ptm->tm_year+1900;

  if( ((currentMonth >= 2) && (monthDay >= 13)) || ((currentMonth <= 11) && (monthDay < 6)) ){
    return true;
  }
  else{
    return false;
  }
 
}
 
 
int timeMinCalc(){
  return 60*getTime(0)+ (12*60*getTime(3) )+ getTime(1);             //calculates current time in terms of minutes
}



